/****************************************************************/
/*                                                              */
/*                           dosfns.c                           */
/*                                                              */
/*                         DOS functions                        */
/*                                                              */
/*                      Copyright (c) 1995                      */
/*                      Pasquale J. Villani                     */
/*                      All Rights Reserved                     */
/*                                                              */
/* This file is part of DOS-C.                                  */
/*                                                              */
/* DOS-C is free software; you can redistribute it and/or       */
/* modify it under the terms of the GNU General Public License  */
/* as published by the Free Software Foundation; either version */
/* 2, or (at your option) any later version.                    */
/*                                                              */
/* DOS-C is distributed in the hope that it will be useful, but */
/* WITHOUT ANY WARRANTY; without even the implied warranty of   */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See    */
/* the GNU General Public License for more details.             */
/*                                                              */
/* You should have received a copy of the GNU General Public    */
/* License along with DOS-C; see the file COPYING.  If not,     */
/* write to the Free Software Foundation, 675 Mass Ave,         */
/* Cambridge, MA 02139, USA.                                    */
/****************************************************************/

#include "portab.h"

#ifdef VERSION_STRINGS
static BYTE *dosfnsRcsId =
    "$Id: dosfns.c 1564 2011-04-08 18:27:48Z bartoldeman $";
#endif

#include "globals.h"

STATIC int do_remote_lock_unlock(__FAR(sft)sftp,    /* SFT for file */
                             unsigned long ofs, /* offset into file */
                             unsigned long len, /* length (in bytes) of region to lock or unlock */
                             int unlock);       /* one to unlock; zero to lock */

__FAR(struct cds)get_cds_unvalidated(unsigned drive)
{
  if (drive >= lastdrive)
    return NULL;
  return &CDSp[drive];
}

/* get current directory structure for drive
   return NULL if the CDS is not valid or the
   drive is not within range */
__FAR(struct cds)get_cds(unsigned drive)
{
  __FAR(struct cds)cdsp;
  unsigned flags;

  if (drive >= lastdrive)
    return NULL;
  cdsp = &CDSp[drive];
  flags = cdsp->cdsFlags;
  /* Entry is disabled or JOINed drives are accessable by the path only */
  if (!(flags & CDSVALID) || (flags & CDSJOINED) != 0)
    return NULL;
  if (!(flags & CDSNETWDRV) && cdsp->cdsDpb == NULL)
    return NULL;
  return cdsp;
}

/* same, but on input drv is 0 for default or 1=A, 2=B, etc. */
__FAR(struct cds)get_cds1(unsigned drv)
{
  if (drv-- == 0) /* get default drive or convert to 0 = A:, 1 = B:, ... */
    drv = default_drive;
  return get_cds(drv);
}

#ifdef WITHFAT32
__FAR(struct dpb) GetDriveDPB(UBYTE drive, COUNT * rc)
{
  __FAR(struct dpb)dpb;
  __FAR(struct cds)cdsp;

  cdsp = get_cds1(drive);

  if (cdsp == NULL)
  {
    *rc = DE_INVLDDRV;
    return NULL;
  }
  dpb = cdsp->cdsDpb;
  if (dpb == NULL || cdsp->cdsFlags & CDSNETWDRV)
  {
    *rc = DE_INVLDDRV;
    return NULL;
  }

  *rc = SUCCESS;
  return dpb;
}
#endif

int idx_to_sft_(int SftIndex)
{
  /*called from below and int2f/ax=1216*/
  __FAR(sfttbl)sp;

  lpCurSft = (__FAR(sft)) - 1;
  if (SftIndex < 0)
    return -1;

  /* Get the SFT block that contains the SFT      */
  for (sp = sfthead; !SFTTBL_END(sp); sp = sp->sftt_next)
  {
    if (SftIndex < sp->sftt_count)
    {
      /* finally, point to the right entry            */
      lpCurSft = (__FAR(sft)) & (sp->sftt_table[SftIndex]);
      return SftIndex;
    }
    SftIndex -= sp->sftt_count;
  }

  /* If not found, return an error                */
  return -1;
}

__FAR(sft) idx_to_sft(int SftIndex)
{
  /* called internally only */
  SftIndex = idx_to_sft_(SftIndex);
  /* if not opened, the SFT is useless            */
  if (SftIndex == -1 || lpCurSft->sft_count == 0)
    return (__FAR(sft)) - 1;
  return lpCurSft;
}

int get_sft_idx(UCOUNT hndl)
{
  __FAR(psp)p = MK_FP(cu_psp, 0);
  int idx;

  if (hndl >= p->ps_maxfiles)
    return DE_INVLDHNDL;

  idx = p->ps_filetab[hndl];
  return idx == 0xff ? DE_INVLDHNDL : idx;
}

__FAR(sft)get_sft(UCOUNT hndl)
{
  /* Get the SFT block that contains the SFT      */
  return idx_to_sft(get_sft_idx(hndl));
}

DWORD DosReadSftExt(int sft_idx, size_t n, __FAR(void)bp, BOOL allow_echo,
    BOOL check_break)
{
  /* Get the SFT block that contains the SFT      */
  __FAR(sft)s = idx_to_sft(sft_idx);

  if (FP_OFF(s) == (UWORD) - 1)
  {
    return DE_INVLDHNDL;
  }
  /* If for read and write-only or for write and read-only then exit */
  if (s->sft_mode & _O_WRONLY)
  {
    return DE_ACCESS;
  }

/*
 *   Do remote first or return error.
 *   must have been opened from remote.
 */
  if (s->sft_flags & SFT_FSHARED)
  {
    DWORD XferCount;
    __FAR(VOID)save_dta;

    save_dta = dta;
    lpCurSft = s;
    current_filepos = s->sft_posit;     /* needed for MSCDEX */
    dta = bp;
    XferCount = remote_rw(REM_READ, s, n);
    dta = save_dta;
    return XferCount;
  }

  /* Do a device transfer if device                   */
  if (s->sft_flags & SFT_FDEVICE)
  {
    __FAR(struct dhdr)dev = s->sft_dev;

    /* Now handle raw and cooked modes      */
    if (s->sft_flags & SFT_FBINARY)
    {
      DWORD rc = BinaryCharIO(dev, n, bp, C_INPUT);
      return rc;
    }
    else
    /* cooked mode */
    {
      DWORD rc;

      /* Test for eof and exit                */
      /* immediately if it is                 */
      if (!(s->sft_flags & SFT_FEOF))
        return 0;

      if (s->sft_flags & SFT_FCONIN && allow_echo)
        rc = read_line_handle(sft_idx, n, bp, check_break);
      else
        rc = cooked_read(dev, n, bp, check_break);
      if (*(__FAR(char))bp == CTL_Z)
        s->sft_flags &= ~SFT_FEOF;
      return rc;
    }
  }

  /* a block transfer                           */
  /* /// Added for SHARE - Ron Cemer */
  if (IsShareInstalled(FALSE) && (s->sft_shroff >= 0))
  {
    int rc = share_access_check(s->sft_shroff, s->sft_posit,
                                 (UDWORD)n, s->sft_dcb->dpb_device,
                                 0x3f);
    if (rc != SUCCESS)
      return rc;
  }
  /* /// End of additions for SHARE - Ron Cemer */
  return rwblock(sft_idx, bp, n, XFR_READ);
}

DWORD DosReadSft(int sft_idx, size_t n, __FAR(void)bp)
{
  return DosReadSftExt(sft_idx, n, bp, TRUE, break_ena);
}

unsigned char read_char(int sft_in, BOOL check_break)
{
  unsigned char c;
  DWORD rc = DosReadSftExt(sft_in, 1, MK_FAR_SCP(c), FALSE, check_break);
  return (rc == 1 ? c : 0);
}

unsigned char read_char_stdin(BOOL check_break)
{
  return read_char(get_sft_idx(STDIN), check_break);
}

DWORD DosWriteSftUnchecked(int sft_idx, size_t n, __XFAR(void)bp)
{
  /* Get the SFT block that contains the SFT      */
  __FAR(sft)s = idx_to_sft(sft_idx);

  if (FP_OFF(s) == (UWORD) - 1)
  {
    return DE_INVLDHNDL;
  }

/*
 *   Do remote first or return error.
 *   must have been opened from remote.
 */
  if (s->sft_flags & SFT_FSHARED)
  {
    DWORD XferCount;
    __FAR(VOID)save_dta;

    save_dta = dta;
    lpCurSft = s;
    current_filepos = s->sft_posit;     /* needed for MSCDEX */
    dta = bp;
    XferCount = remote_rw(REM_WRITE, s, n);
    dta = save_dta;
    return XferCount;
  }

  /* Do a device transfer if device                   */
  if (s->sft_flags & SFT_FDEVICE)
  {
    __FAR(struct dhdr)dev = s->sft_dev;

    /* Now handle raw and cooked modes      */
    if (s->sft_flags & SFT_FBINARY)
    {
      DWORD rc = BinaryCharIO(dev, n, bp, C_OUTPUT);
      if (rc > 0 && (s->sft_flags & SFT_FCONOUT))
      {
        size_t cnt = (size_t)rc;
        __FAR(const char)p = bp;
        while (cnt--)
          update_scr_pos(*p++, 1);
      }
      return rc;
    }
    else
    /* cooked mode */
    {
      /* reset EOF state (set to no EOF)      */
      s->sft_flags |= SFT_FEOF;

      /* if null just report full transfer    */
      if (s->sft_flags & SFT_FNUL)
        return n;
      else
        return cooked_write(dev, n, bp);
    }
  }

  /* a block transfer                           */
  /* /// Added for SHARE - Ron Cemer */
  if (IsShareInstalled(FALSE) && (s->sft_shroff >= 0))
  {
    int rc = share_access_check(s->sft_shroff, s->sft_posit,
                                 (UDWORD)n, s->sft_dcb->dpb_device,
                                 0x40);
    if (rc != SUCCESS)
      return rc;
  }
  /* /// End of additions for SHARE - Ron Cemer */
  return rwblock(sft_idx, bp, n, XFR_WRITE);
}

DWORD DosWriteSft(int sft_idx, size_t n, __XFAR(void)bp)
{
  __FAR(sft)s = idx_to_sft(sft_idx);
  /* If for read and write-only or for write and read-only then exit */
  if ((s->sft_mode & _O_ACCMODE) == _O_RDONLY)
  {
    return DE_ACCESS;
  }
  return DosWriteSftUnchecked(sft_idx, n, bp);
}

static COUNT SftSeek2(int sft_idx, LONG new_pos, unsigned mode, UDWORD * p_result);
COUNT SftSeek(int sft_idx, LONG new_pos, unsigned mode)
{
  UDWORD result;
  return SftSeek2(sft_idx, new_pos, mode, &result);
}

static unsigned try_long_compat_remote_seek(__FAR(sft) s, LONG new_pos,
	unsigned mode, UDWORD * p_result)
{
  struct seek_struct_S {
    UDWORD member[2];
  };
  struct seek_struct_S seek_struct;
  __FAR(struct seek_struct_S) ssp;
  iregs regs = {};

  if (0 == (s->sft_flags & SFT_FSHARED))
    return 0;

  seek_struct.member[0] = new_pos;
  if (mode == SEEK_SET)
    seek_struct.member[1] = 0;
  else if (0 == (new_pos & 0x80000000))
    seek_struct.member[1] = 0;
  else
    seek_struct.member[1] = 0xFFFFffff;

  regs.es = FP_SEG(s);
  regs.di = FP_OFF(s);
  ssp = MK_FAR(seek_struct);
  regs.ds = FP_SEG_OBJ(&regs, ssp);
  regs.d.x = FP_OFF_OBJ(&regs, ssp);
  regs.c.x = mode;
  regs.a.x = 0x1142 | 0x80;
  regs.flags |= FLG_CARRY;
  call_intr(0x2f, MK_FAR_SCP(regs));
  if (regs.flags & FLG_CARRY)
    return 0;
  if (seek_struct.member[1] == 0 || seek_struct.member[1] == 0xFFFFffff)
  {
    *p_result = seek_struct.member[0];
  }
  else
  {
    *p_result = -1;
  }
  return 1;
}

static COUNT SftSeek2(int sft_idx, LONG new_pos, unsigned mode, UDWORD * p_result)
{
  __FAR(sft)s = idx_to_sft(sft_idx);
  if (FP_OFF(s) == (UWORD) -1)
    return DE_INVLDHNDL;

  /* Test for invalid mode                        */
  if (mode > SEEK_END)
    return DE_INVLDFUNC;

  lpCurSft = s;

  /* Do special return for character devices      */
  if (s->sft_flags & SFT_FDEVICE)
  {
    new_pos = 0;
  }
  else if (try_long_compat_remote_seek(s, new_pos, mode, p_result))
  {
    return SUCCESS;
  }
  else if (mode == SEEK_CUR)
  {
    new_pos += s->sft_posit;
  }
  else if (mode == SEEK_END) /* seek from end of file */
  {
/*
 *  RB list has it as Note:
 *  this function is called by the DOS 3.1+ kernel, but only when seeking
 *  from the end of a file opened with sharing modes set in such a manner
 *  that another process is able to change the size of the file while it
 *  is already open
 */
    if (s->sft_flags & SFT_FSHARED)
      new_pos = remote_lseek(s, new_pos);
    else
      new_pos += s->sft_size;
  }

  s->sft_posit = new_pos;
  *p_result = new_pos;
  return SUCCESS;
}

ULONG DosSeek(unsigned hndl, LONG new_pos, COUNT mode, COUNT *rc)
{
  int sft_idx = get_sft_idx(hndl);
  UDWORD result;

  /* Get the SFT block that contains the SFT      */
  *rc = SftSeek2(sft_idx, new_pos, mode, &result);
  if (*rc == SUCCESS)
    return result;
  return *rc;
}

STATIC long get_free_hndl(void)
{
  __FAR(psp)p = MK_FP(cu_psp, 0);
  __FAR(UBYTE)q = p->ps_filetab;
  __FAR(UBYTE)r = fmemchr(q, 0xff, p->ps_maxfiles);
  if (FP_SEG(r) == 0) return DE_TOOMANY;
  return (unsigned)(r - q);
}

STATIC __FAR(sft)get_free_sft(COUNT * sft_idx)
{
  COUNT sys_idx = 0;
  __FAR(sfttbl)sp;

  *sft_idx = 0;
  /* Get the SFT block that contains the SFT      */
  for (sp = sfthead; !SFTTBL_END(sp); sp = sp->sftt_next)
  {
    REG COUNT i = sp->sftt_count;
    __FAR(sft)sfti = sp->sftt_table;

    for (; --i >= 0; sys_idx++, sfti++)
    {
      if (sfti->sft_count == 0)
      {
        *sft_idx = sys_idx;

        /* MS NET uses this on open/creat TE */
        {
          current_sft_idx = sys_idx;
        }

        return sfti;
      }
    }
  }
  /* If not found, return an error                */
  return (__FAR(sft)) - 1;
}

__FAR(const char)get_root(__XFAR(const char)fname)
{
  /* find the end                                 */
  REG unsigned length = fstrlen(fname);
  char c;

  /* now back up to first path seperator or start */
  fname += length;
  while (length)
  {
    length--;
    c = *--fname;
    if (c == '/' || c == '\\' || c == ':') {
      fname++;
      break;
    }
  }
  return fname;
}

STATIC void ConvertPathNameToFCBName(char *FCBName, __FAR(const char)PathName)
{
  ConvertNameSZToName83(FCBName, get_root(PathName));
  FCBName[FNAME_SIZE + FEXT_SIZE] = '\0';
}

STATIC void set_fcbname(void)
{
  ConvertPathNameToFCBName(DirEntBuffer.dir_name, PriPathName);
}

/* initialize SFT fields (for open/creat) for character devices */
STATIC int DeviceOpenSft(__FAR(struct dhdr)dhp, __FAR(sft)sftp)
{
  int i;

  sftp->sft_shroff = -1;      /* /// Added for SHARE - Ron Cemer */
  sftp->sft_count += 1;
  sftp->sft_flags =
    (dhp->dh_attr & ~(SFT_MASK | SFT_FSHARED)) | SFT_FDEVICE | SFT_FEOF;
  fmemcpy(sftp->sft_name, dhp->dh_name, FNAME_SIZE);

  /* pad with spaces */
  for (i = FNAME_SIZE + FEXT_SIZE - 1; sftp->sft_name[i] == '\0'; i--)
    sftp->sft_name[i] = ' ';
  /* and uppercase */
  DosUpFMem(sftp->sft_name, FNAME_SIZE + FEXT_SIZE);

  sftp->sft_dev = dhp;
  sftp->sft_date = dos_getdate();
  sftp->sft_time = dos_gettime();
  sftp->sft_attrib = D_DEVICE;

  if (sftp->sft_dev->dh_attr & SFT_FOCRM)
  {
    /* if Open/Close/RM bit in driver's attribute is set
     * then issue an Open request to the driver
     */
    __FAR(struct dhdr)dev = sftp->sft_dev;
    if (BinaryCharIO(dev, 0, MK_FP(0x0000, 0x0000), C_OPEN) != SUCCESS)
      return DE_ACCESS;
  }
  return SUCCESS;
}

/*
extended open codes
0000 0000 always fail
0000 0001 open O_OPEN
0000 0010 replace O_TRUNC

0001 0000 create new file O_CREAT
0001 0001 create if not exists, open if exists O_CREAT | O_OPEN
0001 0010 create O_CREAT | O_TRUNC

bits for flags (bits 11-8 are internal FreeDOS bits only)
15 O_FCB  called from FCB open
14 O_SYNC commit for each write (not implemented yet)
13 O_NOCRIT do not invoke int23 (not implemented yet)
12 O_LARGEFILE allow files >= 2gb but < 4gb (not implemented yet)
11 O_LEGACY not called from int21/ah=6c: find right fn for redirector
10 O_CREAT if file does not exist, create it
9 O_TRUNC if file exists, truncate and open it \ not both
8 O_OPEN  if file exists, open it              /
7 O_NOINHERIT do not inherit handle on exec
6 \
5  - sharing modes
4 /
3 reserved
2 bits 2,1,0 = 100: RDONLY and do not modify file's last access time
                    (not implemented yet)
1 \ 0=O_RDONLY, 1=O_WRONLY,
0 / 2=O_RDWR, 3=O_EXECCASE (preserve case for redirector EXEC,
                            (not implemented yet))
*/

long DosOpenSft(__FAR(const char) fname, unsigned flags, unsigned attrib)
{
  COUNT sft_idx;
  __FAR(sft)sftp;
  __FAR(struct dhdr)dhp;
  long result;

  result = truename(fname, PriPathName, CDS_MODE_CHECK_DEV_PATH);
  if (result < SUCCESS)
    return result;

  set_fcbname();

  /* now get a free system file table entry       */
  if ((sftp = get_free_sft(&sft_idx)) == (__FAR(sft)) - 1)
    return DE_TOOMANY;

  fmemset(sftp, 0, sizeof(sft));

  sftp->sft_psp = cu_psp;
  sftp->sft_mode = flags & 0xf0ff;
  OpenMode = (BYTE) flags;

  sftp->sft_shroff = -1;        /* /// Added for SHARE - Ron Cemer */
  sftp->sft_attrib = attrib = attrib | D_ARCHIVE;

  if (result & IS_NETWORK)
  {
    int status;
    unsigned cmd;
    if ((flags & (_O_TRUNC | _O_CREAT)) == _O_CREAT)
      attrib |= 0x100;

    lpCurSft = sftp;
    cmd = REM_CREATE;
    if (!(flags & _O_LEGACY))
    {
      ext_open_mode = flags & 0x70ff;
      ext_open_attrib = attrib & 0xff;
      ext_open_action = ((flags & 0x0300) >> 8) | ((flags & _O_CREAT) >> 6);
      cmd = REM_EXTOC;
    }
    else if (!(flags & _O_CREAT))
    {
      cmd = REM_OPEN;
      attrib = (BYTE)flags;
    }
    status = (WORD)network_redirector_mx(cmd, sftp, attrib);
    if (status >= SUCCESS)
    {
      if (sftp->sft_count == 0)
        sftp->sft_count++;
      return sft_idx | ((long)status << 16);
    }
    return status;
  }

  /* check for a device   */
  if ((result & IS_DEVICE) && (dhp = IsDevice(fname)) != NULL)
  {
    int rc = DeviceOpenSft(dhp, sftp);
    /* check the status code returned by the
     * driver when we tried to open it
     */
    if (rc < SUCCESS)
      return rc;
    return sft_idx;
  }

  /* First test the flags to see if the user has passed a valid   */
  /* file mode...                                                 */
  if ((flags & _O_ACCMODE) > 2)
    return DE_INVLDACC;

  /* NEVER EVER allow directories to be created */
  /* ... though FCBs are weird :) */
  if (!(flags & _O_FCB) &&
      (attrib & ~(D_RDONLY | D_HIDDEN | D_SYSTEM | D_ARCHIVE | D_VOLID)))
    return DE_ACCESS;

/* /// Added for SHARE.  - Ron Cemer */
  if (IsShareInstalled(TRUE))
  {
    __FAR(struct cds)cdsp = get_cds(PriPathName[0] - 'A');
    if (!cdsp || !cdsp->cdsDpb)
      return DE_ACCESS;
    if ((sftp->sft_shroff =
         share_open_file(PriPathName, cu_psp,
                          flags & 0x03, (flags >> 4) & 0x07,
                          !!(attrib & D_RDONLY),
                          cdsp->cdsDpb->dpb_device, 0x3d)) < 0)
      return sftp->sft_shroff;
  }

/* /// End of additions for SHARE.  - Ron Cemer */

  sftp->sft_count++;
  sftp->sft_flags = PriPathName[0] - 'A';
  result = dos_open(PriPathName, flags, attrib, sft_idx);
  if (result < 0)
  {
/* /// Added for SHARE *** CURLY BRACES ADDED ALSO!!! ***.  - Ron Cemer */
    if (IsShareInstalled(TRUE))
    {
      share_close_file(sftp->sft_shroff);
      sftp->sft_shroff = -1;
    }
/* /// End of additions for SHARE.  - Ron Cemer */
    sftp->sft_count--;
    return result;
  }
  return sft_idx | ((long)result << 16);
}

long DosOpen(__FAR(char) fname, unsigned mode, unsigned attrib)
{
  long result;
  unsigned hndl;

  /* test if mode is in range                     */
  if ((mode & ~O_VALIDMASK) != 0)
    return DE_INVLDACC;

  /* get a free handle  */
  if ((result = get_free_hndl()) < 0)
    return result;
  hndl = (unsigned)result;

  result = DosOpenSft(fname, mode, attrib);
  if (result < SUCCESS)
    return result;

  ((__FAR(psp))MK_FP(cu_psp, 0))->ps_filetab[hndl] = (UBYTE)result;
  return hndl | (result & 0xffff0000l);
}

COUNT CloneHandle(unsigned hndl)
{
  /* now get the system file table entry                          */
  __FAR(sft)sftp = get_sft(hndl);

  if (sftp == (__FAR(sft)) -1 || (sftp->sft_mode & _O_NOINHERIT))
    return DE_INVLDHNDL;

  /* now that we have the system file table entry, get the fnode  */
  /* index, and increment the count, so that we've effectively    */
  /* cloned the file.                                             */
  sftp->sft_count += 1;
  return SUCCESS;
}

long DosDup(unsigned Handle)
{
  long NewHandle;

  if ((NewHandle = get_free_hndl()) < 0)
    return NewHandle;

  if (DosForceDup(Handle, (unsigned)NewHandle) < 0)
    return DE_INVLDHNDL;
  else
    return NewHandle;
}

COUNT DosForceDup(unsigned OldHandle, unsigned NewHandle)
{
  __FAR(psp)p = MK_FP(cu_psp, 0);
  __FAR(sft)Sftp;

  /* Get the SFT block that contains the SFT                      */
  if ((Sftp = get_sft(OldHandle)) == (__FAR(sft)) - 1)
    return DE_INVLDHNDL;

  /* now close the new handle if it's open                        */
  if ((UBYTE) p->ps_filetab[NewHandle] != 0xff)
  {
    COUNT ret;

    if ((ret = DosClose(NewHandle)) != SUCCESS)
      return ret;
  }

  /* If everything looks ok, bump it up.                          */
  p->ps_filetab[NewHandle] = p->ps_filetab[OldHandle];
  /* possible hazard: integer overflow ska*/
  Sftp->sft_count += 1;
  return SUCCESS;
}

COUNT DosCloseSft(int sft_idx, BOOL commitonly)
{
  __FAR(sft)sftp = idx_to_sft(sft_idx);
  int result;

  if (FP_OFF(sftp) == (UWORD) - 1)
    return DE_INVLDHNDL;

  lpCurSft = sftp;
/*
   remote sub sft_count.
 */
  if (sftp->sft_flags & SFT_FSHARED)
  {
    /* _printf("closing SFT %d = %P\n",sft_idx,GET_FP32(sftp)); */
    return network_redirector_fp(commitonly ? REM_FLUSH: REM_CLOSE, sftp);
  }

  if (sftp->sft_flags & SFT_FDEVICE)
  {
    if (sftp->sft_dev->dh_attr & SFT_FOCRM)
    {
      /* if Open/Close/RM bit in driver's attribute is set
       * then issue a Close request to the driver
       */
      __FAR(struct dhdr)dev = sftp->sft_dev;
      if (BinaryCharIO(dev, 0, MK_FP(0x0000, 0x0000), C_CLOSE) != SUCCESS)
        return DE_INVLDHNDL;
    }
    /* now just drop the count if a device */
    if (!commitonly)
      sftp->sft_count -= 1;
    return SUCCESS;
  }

  /* else call file system handler                     */
  result = dos_close(sft_idx);
  if (commitonly || result != SUCCESS)
    return result;

/* /// Added for SHARE *** CURLY BRACES ADDED ALSO!!! ***.  - Ron Cemer */
  if (sftp->sft_count == 1 && IsShareInstalled(TRUE))
  {
    if (sftp->sft_shroff >= 0)
      share_close_file(sftp->sft_shroff);
    sftp->sft_shroff = -1;
  }
/* /// End of additions for SHARE.  - Ron Cemer */
  sftp->sft_count -= 1;
  return SUCCESS;
}

COUNT DosClose(COUNT hndl)
{
  __FAR(psp)p = MK_FP(cu_psp, 0);
  int sft_idx = get_sft_idx(hndl);

  if (FP_OFF(idx_to_sft(sft_idx)) == (UWORD) - 1)
    return DE_INVLDHNDL;

  /* We must close the (valid) file handle before any critical error */
  /* may occur, else e.g. ABORT will try to close the file twice,    */
  /* the second time after stdout is already closed */
  p->ps_filetab[hndl] = 0xff;

  /* Get the SFT block that contains the SFT      */
  return DosCloseSft(sft_idx, FALSE);
}

UWORD DosGetFree(UBYTE drive, UWORD * navc, UWORD * bps, UWORD * nc)
{
  /* navc==NULL means: called from FatGetDrvData, fcbfns.c */
  __FAR(struct dpb)dpbp;
  __FAR(struct cds)cdsp;
  UCOUNT rg[4];
  UWORD spc;

  /* first check for valid drive          */
  spc = -1;
  cdsp = get_cds1(drive);

  if (cdsp == NULL)
    return spc;

  if (cdsp->cdsFlags & CDSNETWDRV)
  {
    if (remote_getfree(cdsp, rg) != SUCCESS)
      return spc;

    /* for int21/ah=1c:
       Undoc DOS says, its not supported for
       network drives. so it's probably OK */
    /* some programs such as RHIDE want it though and
       the redirector can provide all info
       - Bart, 2002 Apr 1 */

    spc = rg[0];
    if (navc != NULL)
    {
      *navc = (COUNT) rg[3];
      spc &= 0xff; /* zero out media ID byte */
    }

    *nc = (COUNT) rg[1];
    *bps = (COUNT) rg[2];
    return spc;
  }

  dpbp = cdsp->cdsDpb;
  if (dpbp == NULL)
    return spc;

  if (navc == NULL)
  {
      /* hazard: no error checking! */
    flush_buffers(dpbp->dpb_unit);
    dpbp->dpb_flags = M_CHANGED;
  }

  if (media_check(dpbp) < 0)
    return spc;
  /* get the data available from dpb      */
  spc = (dpbp->dpb_clsmask + 1);
  *bps = dpbp->dpb_secsize;

  /* now tell fs to give us free cluster  */
  /* count                                */
#ifdef WITHFAT32
  if (ISFAT32(dpbp))
  {
    ULONG cluster_size, ntotal, nfree = 0;

    /* we shift ntotal until it is equal to or below 0xfff6 */
    cluster_size = (ULONG) dpbp->dpb_secsize << dpbp->dpb_shftcnt;
    ntotal = dpbp->dpb_xsize - 1;
    if (navc != NULL)
      nfree = dos_free(dpbp);
    while (ntotal > FAT_MAGIC16 && cluster_size < 0x8000)
    {
      cluster_size <<= 1;
      spc <<= 1;
      ntotal >>= 1;
      nfree >>= 1;
    }
    /* get the data available from dpb      */
    *nc = ntotal > FAT_MAGIC16 ? FAT_MAGIC16 : (UCOUNT) ntotal;

    /* now tell fs to give us free cluster  */
    /* count                                */
    if (navc != NULL)
      *navc = nfree > FAT_MAGIC16 ? FAT_MAGIC16 : (UCOUNT) nfree;
    return spc;
  }
#endif
  /* a passed navc of NULL means: skip free; see FatGetDrvData
     fcbfns.c */
  if (navc != NULL)
    *navc = (COUNT) dos_free(dpbp);
  *nc = dpbp->dpb_size - 1;
  if (spc > 64)
  {
    /* fake for 64k clusters do confuse some DOS programs, but let
       others work without overflowing */
    spc >>= 1;
    if (navc != NULL)
      *navc = ((unsigned)*navc < FAT_MAGIC16 / 2) ?
        ((unsigned)*navc << 1) : FAT_MAGIC16;
    *nc = ((unsigned)*nc < FAT_MAGIC16 / 2) ? ((unsigned)*nc << 1) : FAT_MAGIC16;
  }
  return spc;
}

#ifdef WITHFAT32
#define IS_SLASH(ch) (ch == '\\' || ch == '/')
COUNT DosGetExtFree(__FAR(BYTE) DriveString, __FAR(struct xfreespace) xfsp)
{
  __FAR(struct dpb)dpbp;
  __FAR(struct cds)cdsp;
  UCOUNT rg[5];

  /* ensure all fields known value - clear reserved bytes & set xfs_version.actual to 0 */
  fmemset(xfsp, 0, sizeof(struct xfreespace));
  xfsp->xfs_datasize = sizeof(struct xfreespace);

  /*
    DriveString should be in form of "C:", "C:\", "\", "", ., or .\
    where missing drive is treated as a request for the current drive,
    or network name in form "\\SERVER\share"
    however, network names like \\SERVER\C aren't supported yet
  */
  cdsp = NULL;
  if ( !*DriveString || (*DriveString == '.') || (IS_SLASH(DriveString[0]) && !IS_SLASH(DriveString[1])) )
    cdsp = get_cds(default_drive);  /* if "" or .[\] or \[path] then use current drive */
  else if (DriveString[1] == ':')
    cdsp = get_cds(DosUpFChar(*DriveString) - 'A');  /* assume drive specified */

  if (cdsp == NULL) /* either error, really bad string, or network name */
    return DE_INVLDDRV;

  if (cdsp->cdsFlags & CDSNETWDRV)
  {
    /* Try redirector extension */
    if (remote_getfree_11a3(cdsp, rg) != SUCCESS)
    {
      /* Fallback */
      if (remote_getfree(cdsp, rg) != SUCCESS)
        return DE_INVLDDRV;

      xfsp->xfs_clussize = rg[0];
      xfsp->xfs_totalclusters = rg[1];
      xfsp->xfs_secsize = rg[2];
      xfsp->xfs_freeclusters = rg[3];
    }
    else /* Supports extension */
    {
      UDWORD total, avail;
      UDWORD bps, spc;

      bps = rg[4];
      spc = 1;
      total = ((rg[0] << 16UL) | rg[1]);
      avail = ((rg[2] << 16UL) | rg[3]);

      while (total > 0x00ffffff && spc < 128) {
        spc *= 2;
        avail /= 2;
        total /= 2;
      }
      while (total > 0x00ffffff && bps < 32768) {
        bps *= 2;
        avail /= 2;
        total /= 2;
      }

      xfsp->xfs_secsize = bps;
      xfsp->xfs_clussize = spc;
      xfsp->xfs_totalclusters = total;
      xfsp->xfs_freeclusters = avail;
    }
  }
  else
  {
    dpbp = cdsp->cdsDpb;
    if (dpbp == NULL || media_check(dpbp) < 0)
      return DE_INVLDDRV;
    xfsp->xfs_secsize = dpbp->dpb_secsize;
    xfsp->xfs_totalclusters =
        (ISFAT32(dpbp) ? dpbp->dpb_xsize : dpbp->dpb_size);
    xfsp->xfs_freeclusters = dos_free(dpbp);
    xfsp->xfs_clussize = dpbp->dpb_clsmask + 1;
  }
  xfsp->xfs_totalunits = xfsp->xfs_totalclusters;
  xfsp->xfs_freeunits = xfsp->xfs_freeclusters;
  xfsp->xfs_totalsectors = xfsp->xfs_totalclusters * xfsp->xfs_clussize;
  xfsp->xfs_freesectors = xfsp->xfs_freeclusters * xfsp->xfs_clussize;
  xfsp->xfs_datasize = sizeof(struct xfreespace);

  return SUCCESS;
}
#endif

COUNT DosGetCuDir(UBYTE drive, __FAR(char) s)
{
  char path[3];

  if (drive-- == 0) /* get default drive or convert to 0 = A:, 1 = B:, ... */
    drive = default_drive;
  path[0] = 'A' + (drive & 0x1f);
  path[1] = ':';
  path[2] = '\0';

  if (truename(path, PriPathName, CDS_MODE_SKIP_PHYSICAL) < SUCCESS)
    return DE_INVLDDRV;

  /* skip d:\ */
  fstrcpy(s, PriPathName + 3);
  return SUCCESS;
}

#undef CHDIR_DEBUG
COUNT DosChangeDir(__FAR(const char) s)
{
  COUNT result;

  result = truename(s, PriPathName, CDS_MODE_CHECK_DEV_PATH);
  if (result < SUCCESS)
    return DE_PATHNOTFND;

  set_fcbname();

  if ((FP_OFF(current_ldt) != 0xFFFF) &&
      (strlen(PriPathName) >= sizeof(current_ldt->cdsCurrentPath)))
    return DE_PATHNOTFND;

#if defined(CHDIR_DEBUG)
  DebugPrintf(("Remote Chdir: n='%Fs' p='%Fs\n", s, PriPathName));
#endif
  /* now get fs to change to new          */
  /* directory                            */
  result = (result & IS_NETWORK ? network_redirector(REM_CHDIR) :
            dos_cd(PriPathName));
#if defined(CHDIR_DEBUG)
  DebugPrintf(("status = %04x, new_path='%Fs'\n", result, cdsd->cdsCurrentPath));
#endif
  if (result != SUCCESS)
    return result;
/*
   Copy the path to the current directory
   structure.

        Some redirectors do not write back to the CDS.
        SHSUCdX needs this. jt
*/
  if (FP_OFF(current_ldt) != 0xFFFF)
  {
     fstrcpy(current_ldt->cdsCurrentPath, PriPathName);
     if (PriPathName[7] == 0)
       current_ldt->cdsCurrentPath[8] = 0; /* Need two Zeros at the end */
  }
  return SUCCESS;
}

STATIC int pop_dmp(int rc, __FAR(dmatch) dmp)
{
  dta = dmp;
  if (rc == SUCCESS)
  {
    fmemcpy(dta, &sda_tmp_dm, 21);
    dmp->dm_attr_fnd = (BYTE) SearchDir.dir_attrib;
    dmp->dm_time = SearchDir.dir_time;
    dmp->dm_date = SearchDir.dir_date;
    dmp->dm_size = (LONG) SearchDir.dir_size;
    ConvertName83ToNameSZ(dmp->dm_name, SearchDir.dir_name);
  }
  return rc;
}

COUNT DosFindFirst(UCOUNT attr, __FAR(const char) name)
{
  int rc;
  REG __FAR(dmatch)dmp = dta;

  rc = truename(name, PriPathName,
                CDS_MODE_CHECK_DEV_PATH | CDS_MODE_ALLOW_WILDCARDS);
  if (rc < SUCCESS)
    return rc;

  set_fcbname();

  /* /// Added code here to do matching against device names.
     DOS findfirst will match exact device names if the
     filename portion (excluding the extension) contains
     a valid device name.
     Credits: some of this code was ripped off from truename()
     in newstuff.c.
     - Ron Cemer */

  SAttr = (BYTE) attr;

#if defined(FIND_DEBUG)
  DebugPrintf(("Remote Find: n='%Fs\n", PriPathName));
#endif

  dta = &sda_tmp_dm;
  memset(&sda_tmp_dm, 0, sizeof(dmatch)+sizeof(struct dirent));

  if (rc & IS_NETWORK)
    rc = network_redirector_fp(REM_FINDFIRST, current_ldt);
  else if (rc & IS_DEVICE)
  {
    __FAR(const char)p;
    COUNT i;

    /* make sure the next search fails */
    sda_tmp_dm.dm_entry = 0xffff;
    /* Found a matching device. Hence there cannot be wildcards. */
    SearchDir.dir_attrib = D_DEVICE;
    SearchDir.dir_time = dos_gettime();
    SearchDir.dir_date = dos_getdate();
    p = get_root(PriPathName);
    memset(SearchDir.dir_name, ' ', FNAME_SIZE + FEXT_SIZE);
    for (i = 0; i < FNAME_SIZE && *p && *p != '.'; i++)
      SearchDir.dir_name[i] = *p++;
    rc = SUCCESS;
    /* /// End of additions.  - Ron Cemer ; heavily edited - Bart Oldeman */
  }
  else
    rc = dos_findfirst(attr, PriPathName);

  return pop_dmp(rc, dmp);
}

COUNT DosFindNext(void)
{
  COUNT rc;
  REG __FAR(dmatch)dmp = dta;

/*
 *  The new version of SHSUCDX 1.0 looks at the dm_drive byte to
 *  test 40h. I used RamView to see location MSD 116:04be and
 *  FD f??:04be, the byte set with 0xc4 = Remote/Network drive 4.
 *  Ralf Brown docs for dos 4eh say bit 7 set == remote so what is
 *  bit 6 for?
 *  SHSUCDX Mod info say "test redir not network bit".
 *  Just to confuse the rest, MSCDEX sets bit 5 too.
 *
 *  So, assume bit 6 is redirector and bit 7 is network.
 *  jt
 *  Bart: dm_drive can be the drive _letter_.
 *  but better just stay independent of it: we only use
 *  bit 7 to detect a network drive; the rest untouched.
 *  RBIL says that findnext can only return one error type anyway
 *  (12h, DE_NFILES)
 */
#if 0
  _printf("findnext: %d\n", dmp->dm_drive);
#endif
  fmemcpy(&sda_tmp_dm, dmp, 21);

  /* findnext will always fail on a volume id search or device name */
  if ((sda_tmp_dm.dm_attr_srch & ~(D_RDONLY | D_ARCHIVE | D_DEVICE)) == D_VOLID
      || (!(sda_tmp_dm.dm_drive & 0x80) && sda_tmp_dm.dm_entry == 0xffff))
    return DE_NFILES;

  memset(&SearchDir, 0, sizeof(struct dirent));
  dta = &sda_tmp_dm;
  rc = (sda_tmp_dm.dm_drive & 0x80) ?
    network_redirector_fp(REM_FINDNEXT, &sda_tmp_dm) : dos_findnext();

  return pop_dmp(rc, dmp);
}

COUNT DosGetFtime(COUNT hndl, date * dp, _time * tp)
{
  __FAR(sft)s;

  /* Get the SFT block that contains the SFT      */
  if (FP_OFF(s = get_sft(hndl)) == (UWORD) - 1)
    return DE_INVLDHNDL;

  *dp = s->sft_date;
  *tp = s->sft_time;
  return SUCCESS;
}

COUNT DosSetFtimeSft(int sft_idx, date dp, _time tp)
{
  /* Get the SFT block that contains the SFT      */
  __FAR(sft)s = idx_to_sft(sft_idx);

  if (FP_OFF(s) == (UWORD) - 1)
    return DE_INVLDHNDL;

  /* If SFT entry refers to a device, do nothing */
  if (s->sft_flags & SFT_FDEVICE)
    return SUCCESS;

  s->sft_flags |= SFT_FDATE;
  s->sft_date = dp;
  s->sft_time = tp;

  return SUCCESS;
}

COUNT DosGetFattr(__FAR(const char) name)
{
  COUNT result;

  result = truename(name, PriPathName, CDS_MODE_CHECK_DEV_PATH);
  if (result < SUCCESS)
    return result;

/* /// Added check for "d:\", which returns 0x10 (subdirectory) under DOS.
       - Ron Cemer */
           /* Theoretically: If the redirectory's qualify function
               doesn't return nonsense this check can be reduced to
               PriPathname[3] == 0, because local path names always
               have the three-byte string ?:\ and UNC path shouldn't
               validy consist of just two slashes.
               -- 2001/09/03 ska*/

  if (PriPathName[3] == '\0')
    return 0x10;

  set_fcbname();

  if (result & IS_NETWORK)
    return remote_getfattr();

  if (result & IS_DEVICE)
    return DE_FILENOTFND;

  return dos_getfattr(PriPathName);
}

/* This function is almost identical to DosGetFattr().
   Maybe it is nice to join both functions.
       -- 2001/09/03 ska*/
COUNT DosSetFattr(__FAR(const char) name, UWORD attrib)
{
  COUNT result;

  result = truename(name, PriPathName, CDS_MODE_CHECK_DEV_PATH);
  if (result < SUCCESS)
    return result;

  set_fcbname();

  if (result & IS_NETWORK)
    return remote_setfattr(attrib);

  if (result & IS_DEVICE)
    return DE_FILENOTFND;

  if (IsShareInstalled(TRUE))
  {
    UWORD mode, psp;
    /* SHARE refuses to change attrs if file is opened */
    if (share_is_file_open(PriPathName, &mode, &psp) &&
        (((mode >> 8) != 0 /* SHARE_COMPAT */) || psp != cu_psp))
      return DE_ACCESS;
  }
  return dos_setfattr(PriPathName, attrib);
}

UBYTE DosSelectDrv(UBYTE drv)
{
  current_ldt = get_cds(drv);

  if (current_ldt != NULL)
    default_drive = drv;

  return lastdrive;
}

COUNT DosDelete(__FAR(const char) path, int attrib)
{
  COUNT result;
  UWORD mode, psp;

  result = truename(path, PriPathName, CDS_MODE_CHECK_DEV_PATH);
  if (result < SUCCESS)
    return result;

  set_fcbname();

  if (result & IS_NETWORK)
    return network_redirector(REM_DELETE);

  if (result & IS_DEVICE)
    return DE_FILENOTFND;

  if (IsShareInstalled(TRUE) && share_is_file_open(PriPathName, &mode, &psp) &&
      (((mode >> 8) != 0 /* SHARE_COMPAT */) || psp != cu_psp))
    return DE_ACCESS;

  return dos_delete(PriPathName, attrib);
}

COUNT DosRenameTrue(__FAR(const char) path1, __FAR(const char) path2, int attrib)
{
  UWORD mode, psp;

  if (path1[0] != path2[0])
  {
    return DE_DEVICE; /* not same device */
  }
  if (FP_OFF(current_ldt) == 0xFFFF || (current_ldt->cdsFlags & CDSNETWDRV))
    return network_redirector(REM_RENAME);

  if (IsShareInstalled(TRUE) && share_is_file_open(path1, &mode, &psp) &&
      (((mode >> 8) != 0 /* SHARE_COMPAT */) || psp != cu_psp))
    return DE_ACCESS;

  return dos_rename(path1, path2, attrib);
}

COUNT DosRename(__FAR(const char) path1, __FAR(const char) path2)
{
  COUNT result;

  result = truename(path2, SecPathName, CDS_MODE_CHECK_DEV_PATH);
  if (result < SUCCESS)
    return result;

  if ((result & (IS_NETWORK | IS_DEVICE)) == IS_DEVICE)
    return DE_FILENOTFND;

  result = truename(path1, PriPathName, CDS_MODE_CHECK_DEV_PATH);
  if (result < SUCCESS)
    return result;

  set_fcbname();

  if ((result & (IS_NETWORK | IS_DEVICE)) == IS_DEVICE)
    return DE_FILENOTFND;

  return DosRenameTrue(PriPathName, SecPathName, D_ALL);
}

COUNT DosMkRmdir(__FAR(const char) dir, int action)
{
  COUNT result;

  result = truename(dir, PriPathName, CDS_MODE_CHECK_DEV_PATH);
  if (result < SUCCESS)
    return result;

  set_fcbname();

  if (result & IS_NETWORK)
    return network_redirector(action == 0x39 ? REM_MKDIR : REM_RMDIR);

  if (result & IS_DEVICE)
    return DE_ACCESS;

  return (action == 0x39 ? dos_mkdir : dos_rmdir)(PriPathName);
}

/* /// Added for SHARE.  - Ron Cemer */

COUNT DosLockUnlock(COUNT hndl, LONG pos, LONG len, COUNT unlock)
{
  __FAR(sft)s;

  /* Get the SFT block that contains the SFT      */
  if (FP_OFF(s = get_sft(hndl)) == (UWORD) - 1)
    return DE_INVLDHNDL;

  if (s->sft_flags & SFT_FSHARED)
    return do_remote_lock_unlock(s, pos, len, unlock);

  /* Invalid function unless SHARE is installed or remote. */
  if (!IsShareInstalled(FALSE))
    return DE_INVLDFUNC;

  /* Lock violation if this SFT entry does not support locking. */
  if (s->sft_shroff < 0)
    return DE_LOCK;

  /* Let SHARE do the work. */
  return share_lock_unlock(s->sft_shroff, pos, len, unlock);
}

/* /// End of additions for SHARE.  - Ron Cemer */

/*
 * This seems to work well.
 */

/* check for a device
   returns device header if match, else returns NULL
   can only match character devices (as only they have names)
 */
__FAR(struct dhdr)IsDevice(__XFAR(const char)fname)
{
  __FAR(struct dhdr)dhp;
  __FAR(const char)froot = get_root(fname);
  int i;

/* /// BUG!!! This is absolutely wrong.  A filename of "NUL.LST" must be
       treated EXACTLY the same as a filename of "NUL".  The existence or
       content of the extension is irrelevent in determining whether a
       filename refers to a device.
       - Ron Cemer
  // if we have an extension, can't be a device <--- WRONG.
  if (*froot != '.')
  {
*/

/*  BUGFIX: MSCD000<00> should be handled like MSCD000<20> TE
    ie the 8 character device name may be padded with spaces ' ' or NULs '\0'

    Note: fname is assumed an ASCIIZ string (ie not padded, unknown length)
    but the name in the device header is assumed FNAME_SIZE and padded.  KJD
*/


  /* check for names that will never be devices to avoid checking all device headers.
     only the file name (not path nor extension) need be checked, "" == root or empty name
   */
  if ( (*froot == '\0') ||
       ((*froot=='.') && ((*(froot+1)=='\0') || (*(froot+2)=='\0' && *(froot+1)=='.')))
     )
  {
    return NULL;
  }

  /* cycle through all device headers checking for match */
  for (dhp = (__FAR(struct dhdr))&nul_dev; dhp != (__FAR(struct dhdr))-1;
       dhp = dhp->dh_next)
  {
    if (!(dhp->dh_attr & ATTR_CHAR))  /* if this is block device, skip */
      continue;

    for (i = 0; i < FNAME_SIZE; i++)
    {
      unsigned char c1 = (unsigned char)froot[i];
      /* ignore extensions and handle filenames shorter than FNAME_SIZE */
      if (c1 == '.' || c1 == '\0')
      {
        /* check if remainder of device name consists of spaces or nulls */
        for (; i < FNAME_SIZE; i++)
        {
          unsigned char c2 = dhp->dh_name[i];
          if (c2 != ' ' && c2 != '\0')
            break;
        }
        break;
      }
      if (DosUpFChar(c1) != DosUpFChar(dhp->dh_name[i]))
        break;
    }

    /* if found a match then return device header */
    if (i == FNAME_SIZE)
      return dhp;
  }

  return NULL;
}

/* /// End of additions for SHARE.  - Ron Cemer */

COUNT DosTruename(__FAR(const char)src, __FAR(char)dest)
{
  /* RBIL: The buffer has be unchanged, if the call fails.
     Therefore, the name is created in an internal buffer
     and copied into the user buffer only on success.
  */
  COUNT rc = truename(src, PriPathName, CDS_MODE_ALLOW_WILDCARDS);
  if (rc >= SUCCESS)
  {
    fstrcpy(dest, PriPathName);
    set_fcbname();
  }
  return rc;
}

STATIC int do_remote_lock_unlock(__FAR(sft)sftp,     /* SFT for file */
                              unsigned long ofs, /* offset into file */
                              unsigned long len, /* length (in bytes) of region to lock or unlock */
                              int unlock)
                                 /* one to unlock; zero to lock */
{
  struct remote_lock_unlock param_block;
  param_block.ofs = ofs;
  param_block.len = len;
  return remote_lock_unlock(sftp, unlock, MK_FAR_SCP(param_block));
}
